"""
Emission energy measurements realized at ONERA, Toulouse, France using the
DEESSE test facility on 2018-06-04.

The Ag sample was heated at 200 celsius during 2 hours.

It is called "heated Ag#1" in my PhD :cite:`Placais2021`. Emission energy data
was corrected using M. Villemant procedure :cite:`Villemant2018`.

This data can be loaded with :class:`.pandas_loader.PandasLoader`.

"""
